/*
 * faaconv.c - program to convert National Flight Data Center (NFDC) files
 *
 * Copyright (C) 1998, John C. Peterson <mailto:jaypee@netcom.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2. A copy is included in this distribution in the file
 * named "LICENSE".
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License, version 2, for more details.
 *
 * $Id: faaconv.c,v 1.1 1998/05/07 03:52:28 jcp Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "faaconv.h"

FILE *fpinput = stdin;
FILE *fpoutput = stdout;
char *inputfile = NULL;
char *outputfile = NULL;

db_t db_type;
bool_t option_k;
write_rec_t write_rec;

bool_t model_required = TRUE;
bool_t model_priority = TRUE;
char *model_directory = DEF_MODEL_DIR;
char *model_filename = "wmm";
char *model_fullname;
field_model_t model;

/* -------------------------------------------------------------------------- */

void print_usage(void)
{

  fprintf(stderr, "Usage: faaconv [options] inputfile [outputfile]\n");
  fprintf(stderr, " the following options are recognized:\n");
  fprintf(stderr, " -f arg specify geomagnetic field model, arg is one of:\n");
  fprintf(stderr, "          igrf - Intl Geomagnetic Reference Field\n");
  fprintf(stderr, "          wmm - Dept of Defense World Magnetic Model\n");
  fprintf(stderr, "          (the default value is wmm)\n");
  fprintf(stderr, " -h     print this help message and then exit\n");
  fprintf(stderr, " -k     prefix 3 char alpha airport ids with \"K\"\n");
  fprintf(stderr, " -m arg specify magnetic variation, arg is one of:\n");
  fprintf(stderr, "          db - use database entry (when available)\n");
  fprintf(stderr, "          fm - always use geomagnetic field model\n");
  fprintf(stderr, "          (the default value is fm)\n");
  fprintf(stderr, " -t arg specify output format type, arg is one of:\n");
  fprintf(stderr, "          fplan - generate fplan (nav) database format\n");
  fprintf(stderr, "          generic - generate a simple report-like format\n");
  fprintf(stderr, "          icao - generate ICAO Map world file format\n");
  fprintf(stderr, "          nav - generate nav (fplan) database format\n");
  fprintf(stderr, "          (the default value is generic)\n");

}

/* -------------------------------------------------------------------------- */

int main(int argc, char *argv[])
{

  int nread = 0;
  int nskip = 0;

  int opt_val;
  bool_t opt_err;
  char *suffix;

  int itoken;
  int ntoken;
  char **tok_str;
  char *tok_ptr;
  char *tok_end;
  char c;

  int buf_pos;
  int buf_len;
  char buf[BUF_LEN];
  bool_t parse_err;

  /* write program version to stderr */

  fprintf(stderr, "faaconv: version %s\n", VERSION);

  /* initialize all options with default values */

  write_rec = &write_generic;

  option_k = FALSE;

  /* parse command line options */

  opt_err = FALSE;

  while ((opt_val=getopt(argc,argv,"f:hkm:t:")) != EOF) {

    switch (opt_val) {

      case 'f':
	/* specify field model coefficient file */
        if ((strcmp(optarg,"igrf") == 0) || (strcmp(optarg,"wmm") == 0)) {
          /* save model coefficient file name */
          model_filename = optarg;
        } else {
          fprintf(stderr, "faaconv: unknown model file name: %s\n", optarg);
          opt_err = TRUE;
        }
	break;

      case 'h':
	/* print help message */
        print_usage();
        exit(0);
	break;

      case 'k':
	/* enable prefixing of airport id with "K" */
        option_k = TRUE;
	break;

      case 'm':
	/* specify  convention for magnetic variation */
        if ((strcmp(optarg,"db") == 0) || (strcmp(optarg,"fm") == 0)) {
          model_priority = (strcmp(optarg,"fm") == 0);
        } else {
          fprintf(stderr, "faaconv: magnetic variation option: %s\n", optarg);
          opt_err = TRUE;
        }
	break;

      case 't':
	/* specify output type */
        if (strcmp(optarg,"generic") == 0) {
          /* generic output format */
          write_rec = &write_generic;
        } else if (strcmp(optarg,"icao") == 0) {
          /* ICAO map output format */
          write_rec = &write_icao;
        } else if ((strcmp(optarg,"fplan") == 0) ||
                   (strcmp(optarg,"nav") == 0)) {
          /* fplan (nav) output format */
          write_rec = &write_nav;
        } else {
          fprintf(stderr, "faaconv: unknown output type: %s\n", optarg);
          opt_err = TRUE;
        }
	break;

      case '?':
        /* unexpected flag */
        opt_err = TRUE;
        break;

    }

  }

  /* parse input file argument */

  if ((optind == (argc-2)) || (optind == (argc-1))) {
    /* save inputfile name */
    inputfile = argv[optind++];
    /* attempt to open inputfile */
    if ((fpinput=fopen(inputfile,"r")) == NULL) {
      fprintf(stderr, "faaconv: unable to open inputfile: %s\n", inputfile);
      opt_err = TRUE;
    }
  } else {
    fprintf(stderr, "faaconv: inputfile argument missing\n");
    opt_err = TRUE;
  }

  /* output file argument */

  if (optind == (argc-1)) {
    /* save outputfile name */
    outputfile = argv[optind++];
    /* attempt to open outputfile */
    if ((fpoutput=fopen(outputfile,"w")) == NULL) {
      fprintf(stderr, "faaconv: unable to open outputfile: %s\n", outputfile);
      opt_err = TRUE;
    }
  }

  /* check for option errors */

  if (opt_err) {
    print_usage();
    exit(1);
  }

  /* infer database file type from file suffix */

  suffix = &inputfile[strlen(inputfile)-4];

  if (strncmp(suffix,".apt",sizeof(suffix)) == 0) {
    ntoken = sizeof(apt_t) / sizeof(char *);
    db_type = DB_APT;
  } else if (strncmp(suffix,".fix",sizeof(suffix)) == 0) {
    ntoken = sizeof(fix_t) / sizeof(char *);
    db_type = DB_FIX;
  } else if (strncmp(suffix,".ils",sizeof(suffix)) == 0) {
    ntoken = sizeof(ils_t) / sizeof(char *);
    db_type = DB_ILS;
  } else if (strncmp(suffix,".nav",sizeof(suffix)) == 0) {
    ntoken = sizeof(nav_t) / sizeof(char *);
    db_type = DB_NAV;
  } else if (strncmp(suffix,".rwy",sizeof(suffix)) == 0) {
    ntoken = sizeof(rwy_t) / sizeof(char *);
    db_type = DB_RWY;
  } else {
    fprintf(stderr, "faaconv: inputfile has unknown suffix: %s\n", suffix);
    exit(1);
  }

  /* if it will be required, initialize the geomagnetic field model module */

  if (model_required) {
    struct tm *time_now;
    time_t time_t_now;
    int nbytes;
    /* check for diretory environment variable */
    if (getenv(ENV_MODEL_DIR) != NULL) {
      model_directory = getenv(ENV_MODEL_DIR);
    }
    /* remove any trailing / from directory */
    if (model_directory[strlen(model_directory)-1] == '/') {
      model_directory[strlen(model_directory)-1] = '\0';
    }
    /* allocate memory for full path to file */
    nbytes = strlen(model_directory)+strlen(model_filename)+2;
    if ((model_fullname = (char *)malloc(nbytes)) == NULL) {
      fprintf(stderr, "faaconv: malloc(%d) of model fullname failed\n", nbytes);
      exit(1);
    } else {
      /* form the full name of the model coefficient data file */
      sprintf(model_fullname,"%s/%s", model_directory, model_filename);
    }
    /* initialize the model */
    if (! init_field_model(model_fullname)) {
      fprintf(stderr, "faaconv: init_field_model failed: %s\n",
                      strerror_field_model());
      exit(1);
    }
    /* get current time */
    time(&time_t_now);
    time_now = localtime(&time_t_now);
    model.date = (1900.0L+time_now->tm_year) + (time_now->tm_yday/365.0L);
  }

  /* allocate memory for array of pointers to token strings */

  if ((tok_str = (char **)malloc(ntoken*sizeof(char *))) == NULL) {
    fprintf(stderr, "faaconv: malloc(%d) of token handles failed\n", ntoken);
    exit(1);
  }

  /* read records from input file */

  while (fgets(buf,sizeof(buf),fpinput) != NULL) {

    nread++;

    /* find pointers to all tokens expected in the record */

    buf_pos = 0;
    buf_len = strlen(buf);
    parse_err = FALSE;
    tok_ptr = buf;

    for (itoken = 0; itoken < ntoken; itoken++) {

      /* discard any leading white spaces */

      while ((*tok_ptr) == ' ') tok_ptr++;

      /* save pointer to first non-white char */

      tok_str[itoken] = tok_ptr;

      /* scan for "," seperator, newline char, or end of buffer */

      tok_end = tok_ptr - 1;

      while ((*tok_ptr != ',') && (*tok_ptr != '\n') && (buf_pos < buf_len)) {
        if ((*tok_ptr) != ' ') tok_end = tok_ptr;
        tok_ptr++;
        buf_pos++;
      }

      /* handle different string terminations */

      c = *tok_ptr;

      if (c == ',') {
        /* terminate string */
        *(++tok_end) = '\0';
      } else if (c == '\n') {
        /* should be last token */
        if (itoken < (ntoken-1)) {
          /* missing tokens */
          parse_err = TRUE;
          break;
        } else {
          /* terminate string */
          *(++tok_end) = '\0';
        }
      } else if (buf_pos == buf_len) {
        /* buffer overflow (rare) */
        parse_err = TRUE;
        break;
      }

      /* next token */

      tok_ptr++;
      buf_pos++;

    }

    /* final token should be terminated by '\n' */
/*
**  if (c != '\n') parse_err = TRUE;
*/
    /* convert and write output record */

    if (! parse_err) {
      /* write output record */
      if (! write_rec(tok_str)) nskip++;
    } else {
      nskip++;
    }

  }

  /* write a processing summary to stderr */

  fprintf(stderr, "Read %d records\n", nread);
  fprintf(stderr, "Rejected %d records\n", nskip);

  /* normal exit */

  fclose(fpinput);
  fclose(fpoutput);

  exit(0);

}

/* -------------------------------------------------------------------------- */

